#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE tmGrammar

#include "tmGrammar/Algorithm.hh"
#include "tmGrammar/Function.hh"
#include "tmGrammar/Object.hh"
#include "tmGrammar/Cut.hh"

#include <boost/test/unit_test.hpp>
#include <boost/algorithm/string.hpp>

#include <string>
#include <vector>

BOOST_AUTO_TEST_SUITE(Algorithm)

BOOST_AUTO_TEST_CASE(Algorithm_isGate_valid)
{
  const std::vector<std::string> tokens = {
    Algorithm::AND,
    Algorithm::OR,
    Algorithm::NOT,
    Algorithm::XOR
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Algorithm::isGate(token), true);
  }
}

BOOST_AUTO_TEST_CASE(Algorithm_isGate_invalid)
{
  const std::vector<std::string> tokens = {
    " AND ",
    "MU10",
    "comb{MU0,MU0}",
    "foobar",
    "ANDOR",
    "XORNOT"
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Algorithm::isGate(token), false);
  }
}

BOOST_AUTO_TEST_CASE(Algorithm_parser)
{
  std::vector<std::string> reference;

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("FOO AND NOT BAR"), true);
  // NOTE: RPN (Reverse Polish notation)
  boost::split(reference, "FOO BAR NOT AND", boost::is_any_of(" "));
  BOOST_CHECK_EQUAL_COLLECTIONS(Algorithm::Logic::getTokens().begin(),
                                Algorithm::Logic::getTokens().end(),
                                reference.begin(), reference.end());

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("NOT FOO OR NOT BAR AND BAZ"), true);
  // NOTE: RPN (Reverse Polish notation)
  boost::split(reference, "FOO NOT BAR NOT OR BAZ AND", boost::is_any_of(" "));
  BOOST_CHECK_EQUAL_COLLECTIONS(Algorithm::Logic::getTokens().begin(),
                                Algorithm::Logic::getTokens().end(),
                                reference.begin(), reference.end());

  // functions

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("some_func{FOO42,BAR42,BAZ42}"), true);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("some_func{FOO42,BAR42}[BAZ]"), true);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("another_func{FOO42[BAR],BAR42[BAZ]}[BAZ,FOO]"), true);

  // fault injections

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("AND"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("NOT AND"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("FOO AND"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("AND FOO"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("AND OR AND"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("FOO OR NOT"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("FOO NOT OR BAR"), false);

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Algorithm::parser("NOT FOO AND XOR"), false);
}

BOOST_AUTO_TEST_CASE(Algorithm_Logic)
{
  Algorithm::Logic::clear();
  Algorithm::Logic logic;

  logic.tokens.emplace_back("foo");
  logic.append("bar");
  logic.append("baz");

  std::vector<std::string> reference;
  boost::split(reference, "foo bar baz", boost::is_any_of(" "));
  BOOST_CHECK_EQUAL_COLLECTIONS(logic.getTokens().begin(), logic.getTokens().end(),
                                reference.begin(), reference.end());

  Algorithm::Logic::clear();
  BOOST_CHECK_EQUAL(Logic::getTokens().size(), 0);
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(Object)

BOOST_AUTO_TEST_CASE(Object_parser)
{
  Object::Item item;

  BOOST_CHECK_EQUAL(Object::parser("MU0[MU-UPT_10_20]", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "MU");
  BOOST_CHECK_EQUAL(item.threshold, "0");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.isValidCut("MU-ISO_X"), true);
  BOOST_CHECK_EQUAL(item.isValidCut("JET-ETA_X"), false);
  BOOST_CHECK_EQUAL(item.getObjectName(), "MU0");

  BOOST_CHECK_EQUAL(Object::parser("TAU2p5+1", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "TAU");
  BOOST_CHECK_EQUAL(item.threshold, "2p5");
  BOOST_CHECK_EQUAL(item.bx_offset, "+1");
  BOOST_CHECK_EQUAL(item.isValidCut("TAU-PHI_X"), true);
  BOOST_CHECK_EQUAL(item.isValidCut("MU-ETA_X"), false);
  BOOST_CHECK_EQUAL(item.getObjectName(), "TAU2p5+1");

  BOOST_CHECK_EQUAL(Object::parser("JET123", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "JET");
  BOOST_CHECK_EQUAL(item.threshold, "123");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.isValidCut("JET-SLICE_X"), true);
  BOOST_CHECK_EQUAL(item.isValidCut("MU-ISO_X"), false);
  BOOST_CHECK_EQUAL(item.getObjectName(), "JET123");
}

BOOST_AUTO_TEST_CASE(Object_ext_parser)
{
  Object::Item item;

  BOOST_CHECK_EQUAL(Object::parser("EXT_FOO", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "EXT_FOO");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.getObjectName(), "EXT_FOO");

  BOOST_CHECK_EQUAL(Object::parser("EXT_FOO+0", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "EXT_FOO");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.getObjectName(), "EXT_FOO");

  BOOST_CHECK_EQUAL(Object::parser("EXT_FOO-2", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "EXT_FOO");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "-2");
  BOOST_CHECK_EQUAL(item.getObjectName(), "EXT_FOO-2");

  BOOST_CHECK_EQUAL(Object::parser("EXT_FOO+2", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "EXT_FOO");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+2");
  BOOST_CHECK_EQUAL(item.getObjectName(), "EXT_FOO+2");
}

BOOST_AUTO_TEST_CASE(Object_cent_parser)
{
  Object::Item item;

  BOOST_CHECK_EQUAL(Object::parser("CENT0", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "CENT0");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.getObjectName(), "CENT0");

  BOOST_CHECK_EQUAL(Object::parser("CENT1+0", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "CENT1");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+0");
  BOOST_CHECK_EQUAL(item.getObjectName(), "CENT1");

  BOOST_CHECK_EQUAL(Object::parser("CENT2+2", item), true);
  BOOST_CHECK_EQUAL(item.message, "");
  BOOST_CHECK_EQUAL(item.name, "CENT2");
  BOOST_CHECK_EQUAL(item.threshold, "");
  BOOST_CHECK_EQUAL(item.bx_offset, "+2");
  BOOST_CHECK_EQUAL(item.getObjectName(), "CENT2+2");
}

BOOST_AUTO_TEST_CASE(Object_isObject_true)
{
  const std::vector<std::string> tokens = {
    "MU0[MU-UPT_X]",
    "EG1",
    "TAU2",
    "JET4",
    "TAU2"
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Object::isObject(token), true);
  }
}

BOOST_AUTO_TEST_CASE(Object_isObject_false)
{
  const std::vector<std::string> tokens = {
    "FOO[MU-UPT_X]",
    "BAR",
    " TAU1 ",
    "comb{MU0,MU1}"
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Object::isObject(token), false);
  }
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(Cut)

BOOST_AUTO_TEST_CASE(Cut_parser)
{
  Cut::Item item;
  BOOST_CHECK_EQUAL(Cut::parser("MU-ISO_FOO, MU-ETA_BAR, MU-PHI_BAZ, MU-CHG_POS", item), true);
  BOOST_CHECK_EQUAL(item.message.empty(), true);
  std::vector<std::string> reference;
  boost::split(reference, "MU-ISO_FOO MU-ETA_BAR MU-PHI_BAZ MU-CHG_POS", boost::is_any_of(" "));
  BOOST_CHECK_EQUAL_COLLECTIONS(item.name.begin(), item.name.end(),
                                reference.begin(), reference.end());

  item.name.clear();
  item.message.clear(); // so thats not convenient at all

  BOOST_CHECK_EQUAL(Cut::parser("MU-FOO_BAR", item), false);
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(Function)

BOOST_AUTO_TEST_CASE(Function_parser_true)
{
  const std::vector<std::string> tokens = {
    "comb{MU0,MU0}",
    "dist{TAU0,EG1}[DETA_X,DR_X]",

    "mass_inv{TAU0,EG1}[MASS_X]",
    "mass_inv{TAU0,EG1}[DR_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DETA_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DPHI_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DETA_X,DPHI_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DR_X,DPHI_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DR_X,DETA_X,MASS_X]",
    "mass_inv{TAU0,EG1}[DR_X,DETA_X,DPHI_X,MASS_X]",

    "mass_inv_upt{MU0,MU1}[MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DR_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DETA_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DPHI_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DETA_X,DPHI_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DR_X,DPHI_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DR_X,DETA_X,MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[DR_X,DETA_X,DPHI_X,MASSUPT_X]",

    "mass_inv_dr{TAU0,EG1}[MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DR_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DETA_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DPHI_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DETA_X,DPHI_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DR_X,DPHI_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DR_X,DETA_X,MASSDR_X]",
    "mass_inv_dr{TAU0,EG1}[DR_X,DETA_X,DPHI_X,MASSDR_X]",

    "mass_inv_3{EG0,EG1,EG0}[MASS_X]",
    "mass_inv_3{TAU0,TAU1,TAU2}[MASS_X]",
    "mass_inv_3{MU0,MU1,MU2}[MASS_X]",
    "mass_inv_3{MU0,EG1,EG2}[MASS_X]",
    "mass_inv_3{EG1,MU2,TAU3}[MASS_X]",

    "comb_orm{JET0,TAU0}[ORMDETA_X]",
    "comb_orm{JET0,JET1,TAU0}[ORMDR_X]",
    "comb_orm{JET3,JET2,JET1,TAU0}[ORMDR_X]",
    "comb_orm{JET4,JET3,JET2,JET1,TAU0}[ORMDR_X]",

    "dist_orm{TAU0,EG1}[ORMDETA_X,DR_X]",
    "dist_orm{TAU0,TAU1,EG1}[ORMDETA_X,DR_X]",

    "mass_inv_orm{TAU0,EG1}[ORMDETA_X,MASS_X]",
    "mass_inv_orm{TAU0,TAU1,EG1}[ORMDETA_X,MASS_X]",

    "mass_trv_orm{TAU0,EG1}[ORMDETA_X,MASS_X]",
    "mass_trv_orm{TAU0,TAU1,EG1}[ORMDETA_X,MASS_X]"
  };

  for (const auto& token: tokens)
  {
    Function::Item item;
    BOOST_CHECK_MESSAGE(Function::parser(token, item) == true, token);
    BOOST_CHECK_EQUAL(item.message.empty(), true);
  }
}

BOOST_AUTO_TEST_CASE(Function_parser_false)
{
  const std::vector<std::string> tokens = {
    "comb_orm{JET0}[ORMDETA_X]",
    "comb_orm{JET0,JET0,JET0}[ORMDETA_X]",
    "comb_orm{JET3,JET2}[ORMDETA_X]",
    "comb_orm{JET3,JET2,JET1,JET1}[ORMDETA_X]",
    "comb_orm{JET4,JET3,JET2}[ORMDETA_X]",
    "comb_orm{JET4,JET3,JET2,JET1,JET0}[ORMDR_X]",

    "dist_orm{JET0,JET1}[DR_X]", // missing overlap removal cut
    "dist_orm{JET0,TAU0}[ORMDR_X]", // missing delta cut
    "dist_orm{JET0,TAU0,TAU0}[ORMDR_X,DR_X]",
    "dist_orm{JET0,JET1,JET2,TAU0}[ORMDR_X,DR_X]",
    "dist_orm{JET0,JET1,TAU0,TAU1}[ORMDR_X,DR_X]",

    "mass_inv_3{EG0,EG1}[MASS_X]",
    "mass_inv_3{TAU0,TAU1,TAU2}[MASSDR_X]",

    "mass_inv_orm{JET0,JET1}[MASS_Z]", // missing overlap removal cut
    "mass_inv_orm{JET0,TAU0}[ORMDR_X]", // missing mass cut
    "mass_inv_orm{JET0,TAU0,TAU0}[ORMDR_X,MASS_Z]",
    "mass_inv_orm{JET0,JET1,JET2,TAU0}[ORMDR_X,MASS_Z]",
    "mass_inv_orm{JET0,JET1,TAU0,TAU1}[ORMDR_X,MASS_Z]",

    "mass_trv_orm{JET0,JET1}[MASS_Z]", // missing overlap removal cut
    "mass_trv_orm{JET0,TAU0}[ORMDR_X]", // missing mass cut
    "mass_trv_orm{JET0,TAU0,TAU0}[ORMDR_X,MASS_Z]",
    "mass_trv_orm{JET0,JET1,JET2,TAU0}[ORMDR_X,MASS_Z]",
    "mass_trv_orm{JET0,JET1,TAU0,TAU1}[ORMDR_X,MASS_Z]"
  };

  for (const auto& token: tokens)
  {
    Function::Item item;
    BOOST_CHECK_MESSAGE(Function::parser(token, item) == false, token);
    BOOST_CHECK_EQUAL(item.message.empty(), false);
  }
}

BOOST_AUTO_TEST_CASE(Function_isFunction_true)
{
  const std::vector<std::string> tokens = {
    "comb{MU0,MU0}",
    "dist{TAU0,EG1}[DETA_X,DR_X]",
    "mass{TAU0,EG1}[MASS_X]",
    "mass_inv{TAU0,EG1}[MASS_X]",
    "mass_inv_upt{MU0,MU0}[MASSUPT_X]",
    "mass_inv_dr{MU0,MU1}[MASSDR_X]",
    "comb_orm{JET0,TAU0}[ORMDETA_X]",
    "comb_orm{JET0,JET0,TAU0}[ORMDR_X]",
    "comb_orm{JET3,JET2,JET1,TAU0}[ORMDR_X]",
    "comb_orm{JET4,JET3,JET2,JET1,TAU0}[ORMDR_X]"
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Function::isFunction(token), true);
  }
}

BOOST_AUTO_TEST_CASE(Function_isFunction_false)
{
  const std::vector<std::string> tokens = {
    "mass_inv_upt{JET0,TAU0}[MASSUPT_X]",
    "mass_inv_upt{MU0,MU1}[MASS_X]"
  };

  for (const auto& token: tokens)
  {
    BOOST_CHECK_EQUAL(Function::isFunction(token), false);
  }
}

BOOST_AUTO_TEST_SUITE_END()
